/* 
 * airport.c -- scli airport mode implementation
 *
 * Copyright (C) 2006 Jasper Wallace, <jasper@pointless.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * 
 * @(#) $Id$
 */

#include "scli.h"

#include "if-mib.h"
#include "if-mib-proc.h"

#include "airport-basestation-3-mib.h"
#include <assert.h>

/*
 * TODO:
 *
 * airport monitor wireless
 * airport monitor dhcp
 *
 */

/* caller has to free after use. */
char *mac_to_str(uint8_t *mac)
{
  char *str;
  /* XX:XX:XX:XX:XX:XX\0 */
  str = g_malloc(18);
  snprintf(str, 18, "%02x:%02x:%02x:%02x:%02x:%02x",
    mac[0], mac[1], mac[2], mac[3], mac[4], mac[5]);
  return str;
}

typedef struct {
    guint32 rx;
    guint32 tx;
    guint32 rxerr;
    guint32 txerr;
} ap_wi_stats_t;

static int
show_airport_physical_interfaces_info(scli_interp_t *interp, int argc, char **argv)
{
  airport_basestation_3_mib_physicalInterface_t **apinterfaceTable = NULL;
  int i, code = SCLI_OK;

  g_return_val_if_fail(interp, SCLI_ERROR);
  
  if (argc > 1) {
    return SCLI_SYNTAX_NUMARGS;
  }
  
  if (scli_interp_dry(interp)) {
    return SCLI_OK;
  }
  
  airport_basestation_3_mib_get_physicalInterfacesTable(interp->peer, &apinterfaceTable, 0);
  if (interp->peer->error_status) {
    code = SCLI_SNMP;
    goto done;
  }
  
  if (apinterfaceTable) {
    g_string_sprintfa(interp->header, "NAME SPEED %6s DUPLEX %8s %8s %8s %8s",
                      "STATE", "TX", "RX", "TXERROR", "RXERROR");
    
    for (i = 0; apinterfaceTable[i]; i++) {
      const char *e;

      g_string_sprintfa(interp->result, "%.*s%d ", 
                      (int)apinterfaceTable[i]->_physicalInterfaceNameLength,
                      apinterfaceTable[i]->physicalInterfaceName,
                      *apinterfaceTable[i]->physicalInterfaceUnit);

      g_string_sprintfa(interp->result, "%6s ",
                      fmt_kmg(*apinterfaceTable[i]->physicalInterfaceSpeed));

      e = fmt_enum(airport_basestation_3_mib_enums_physicalInterfaceState,
                  apinterfaceTable[i]->physicalInterfaceState);
      g_string_sprintfa(interp->result, "%s ", e ? e : "");

      e = fmt_enum(airport_basestation_3_mib_enums_physicalInterfaceDuplex,
                  apinterfaceTable[i]->physicalInterfaceDuplex);
      g_string_sprintfa(interp->result, "%6s ", e ? e : "");

      g_string_sprintfa(interp->result, "%8d ",
                      *apinterfaceTable[i]->physicalInterfaceNumTX);
      g_string_sprintfa(interp->result, "%8d ",
                      *apinterfaceTable[i]->physicalInterfaceNumRX);
      g_string_sprintfa(interp->result, "%8d ",
                      *apinterfaceTable[i]->physicalInterfaceNumTXError);
      g_string_sprintfa(interp->result, "%8d",
                      *apinterfaceTable[i]->physicalInterfaceNumRXError);
      g_string_append(interp->result, "\n");
    }
  }
  
  done:
    if (apinterfaceTable) airport_basestation_3_mib_free_physicalInterfacesTable(apinterfaceTable);
    
    
  return code;
}

static int get_wireless(scli_interp_t *interp) {
  airport_basestation_3_mib_wireless_t *wireless = NULL;
  int num = -1;
  
  airport_basestation_3_mib_get_wireless(interp->peer, &wireless, 0);
  
  if (!wireless)
    return -1;
  
  if (wireless->wirelessNumber) {
    num = *(wireless->wirelessNumber);
  }

  airport_basestation_3_mib_free_wireless(wireless);
  
  return num;
}

static int get_dhcp(scli_interp_t *interp) {
  airport_basestation_3_mib_dhcpServer_t *dhcp = NULL;
  int num = -1;
  
  airport_basestation_3_mib_get_dhcpServer(interp->peer, &dhcp, 0);
  
  if (!dhcp)
    return -1;
  
  if (dhcp->dhcpNumber) {
    num = *(dhcp->dhcpNumber);
  }

  airport_basestation_3_mib_free_dhcpServer(dhcp);
  
  return num;
}

static int
show_airport_wireless(scli_interp_t *interp, int argc, char **argv)
{
  airport_basestation_3_mib_wirelessClient_t **apwirelessTable = NULL;
  int i, code = SCLI_OK, num;

  g_return_val_if_fail(interp, SCLI_ERROR);
  
  if (argc > 1) {
    return SCLI_SYNTAX_NUMARGS;
  }
  
  if (scli_interp_dry(interp)) {
    return SCLI_OK;
  }

  num = get_wireless(interp);
  
  if (num == -1) {
    return SCLI_ERROR;
  }

  if (num == 0) {
    g_string_sprintfa(interp->result, "No wireless clients\n");
    return SCLI_OK;
  }
  
  airport_basestation_3_mib_get_wirelessClientsTable(interp->peer, &apwirelessTable, 0);

  if (interp->peer->error_status) {
    code = SCLI_SNMP;
    goto done;
  }
  
  if (apwirelessTable) {
    g_string_sprintfa(interp->header, "%-17s type strength noise rate %8s %8s %8s %8s",
        "MAC", "rx", "tx", "rxerr", "txerr");

    for (i = 0; apwirelessTable[i]; i++) {
      const char *e;
      char *m;

      m = mac_to_str(apwirelessTable[i]->wirelessMacAddress);

      g_string_sprintfa(interp->result, "%s ", m);

      g_free(m);

      e = fmt_enum(airport_basestation_3_mib_enums_wirelessType,
                  apwirelessTable[i]->wirelessType);
      g_string_sprintfa(interp->result, "%4s ", e ? e : "");

/*      g_string_sprintfa(interp->result, "%.*s ", 
                      (int)apwirelessTable[i]->_wirelessDataRatesLength,
                      apwirelessTable[i]->wirelessDataRates);
*/
//      printf("%.*s", (int)apwirelessTable[i]->_wirelessDataRatesLength, apwirelessTable[i]->wirelessDataRates);

      g_string_sprintfa(interp->result, "%8d ",
                      *apwirelessTable[i]->wirelessStrength);

      g_string_sprintfa(interp->result, "%5d ",
                      *apwirelessTable[i]->wirelessNoise);

      g_string_sprintfa(interp->result, "%4d ",
                      *apwirelessTable[i]->wirelessRate);

      g_string_sprintfa(interp->result, "%8d ",
                      *apwirelessTable[i]->wirelessNumTX);
      g_string_sprintfa(interp->result, "%8d ",
                      *apwirelessTable[i]->wirelessNumRX);
      g_string_sprintfa(interp->result, "%8d ",
                      *apwirelessTable[i]->wirelessNumTXErrors);
      g_string_sprintfa(interp->result, "%8d",
                      *apwirelessTable[i]->wirelessNumRXErrors);
      g_string_append(interp->result, "\n");
    }
  } else {
    code = SCLI_ERROR;
  }
  
  done:
    if (apwirelessTable)
      airport_basestation_3_mib_free_wirelessClientsTable(apwirelessTable);
    
  return code;
}

/* used to clean dead entries from the hash table used in monitor_airport_wireless */
gboolean clean_wistats(gpointer key, gpointer value, gpointer data)
{
  ap_wi_stats_t *stats;
  char *mk, *m;
  airport_basestation_3_mib_wirelessClient_t **apwt;
  int i;
  
  mk = key;
  stats = value;
  apwt = data;
  
  for (i = 0; apwt[i]; i++)
  {
    m = mac_to_str(apwt[i]->wirelessMacAddress);
    /* is this key in the table? */
    if (strcmp(m, mk) == 0) {
      g_free(m);
      return FALSE;
    }
    g_free(m);
  }
  /* if we get to here we've not had a match, so the entry is redundant */  
  g_free(key);
  g_free(value);

  return TRUE;
}

static int
monitor_airport_wireless(scli_interp_t *interp, int argc, char **argv)
{
  airport_basestation_3_mib_wirelessClient_t **apwirelessTable = NULL;
  int i, code = SCLI_OK, num;
  static struct timeval last, now;
  double delta;
  static time_t epoch = 0;
  ap_wi_stats_t *stats;
  static GHashTable *mac_to_stats = NULL;

  if (argc > 1) {
    return SCLI_SYNTAX_NUMARGS;
  }

  if (scli_interp_dry(interp)) {
    return SCLI_OK;
  }

  num = get_wireless(interp);
    
  if (num == -1 || num == 0) {
    code = SCLI_ERROR;
    goto done;
  }

  airport_basestation_3_mib_get_wirelessClientsTable(interp->peer, &apwirelessTable, 0);

  if (interp->peer->error_status) {
    code = SCLI_SNMP;
    goto done;
  }
  
  if (!apwirelessTable) {
    code = SCLI_SNMP;
    goto done;
  }
  
  if (epoch < interp->epoch) {
    last.tv_sec = last.tv_usec = 0;
  }

  /*
   * XXX need some kind of atexit callback to free the hashtable.
   */
  if (!mac_to_stats)
  {
    mac_to_stats = g_hash_table_new(g_str_hash, g_str_equal);
  }

  {
    char * m;

    /* go through each table entry seeing if
     * the mac is in the hash
     *
     * if not, add it.
     */
    for (i = 0; apwirelessTable[i]; i++)
    {
      m = mac_to_str(apwirelessTable[i]->wirelessMacAddress);
      
      if (g_hash_table_lookup(mac_to_stats, m) == NULL)
      {
        stats = g_malloc(sizeof(ap_wi_stats_t));
	stats->tx = *(apwirelessTable[i]->wirelessNumTX);
	stats->rx = *(apwirelessTable[i]->wirelessNumRX);
	stats->txerr = *(apwirelessTable[i]->wirelessNumTXErrors);
	stats->rxerr = *(apwirelessTable[i]->wirelessNumRXErrors);

        g_hash_table_insert(mac_to_stats, m, stats);
      } else {
        g_free(m);
      }
    }
    
    /*
     * now loop through the hash keys comparing them to the mac's in
     * the table.
     *
     * if a key is not in the table then ditch it.
     */
    g_hash_table_foreach_remove(mac_to_stats, clean_wistats, apwirelessTable);
  }

  epoch = time(NULL);
  gettimeofday(&now, NULL);
  delta = TV_DIFF(last, now);

  if (apwirelessTable) {
	g_string_sprintfa(interp->header, "%-17s TYPE STRENGTH NOISE RATE %5s %5s %5s %5s",
        "MAC", "RX", "TX", "RXERR", "TXERR");

	for (i = 0; apwirelessTable[i]; i++) {
	    GString *s;
            const char *e;
            char *m;
            
            m = mac_to_str(apwirelessTable[i]->wirelessMacAddress);
            stats = g_hash_table_lookup(mac_to_stats, m);
            
            /*
             * if it's null the key didn't exist, that should never happen.
             */
            assert(stats != NULL);

	    s = interp->result;
	    g_string_sprintfa(s, "%s ", m);
            g_free(m);

            e = fmt_enum(airport_basestation_3_mib_enums_wirelessType,
                        apwirelessTable[i]->wirelessType);
            g_string_sprintfa(interp->result, "%4s ", e ? e : "");

            g_string_sprintfa(interp->result, "%8d ",
                            *apwirelessTable[i]->wirelessStrength);

            g_string_sprintfa(interp->result, "%5d ",
                            *apwirelessTable[i]->wirelessNoise);

            g_string_sprintfa(interp->result, "%4d",
                            *apwirelessTable[i]->wirelessRate);

	    fmt_counter_dt(s, apwirelessTable[i]->wirelessNumTX, &(stats->tx),
			   &last, delta);
	    fmt_counter_dt(s, apwirelessTable[i]->wirelessNumRX, &(stats->rx),
			   &last, delta);
	    fmt_counter_dt(s, apwirelessTable[i]->wirelessNumTXErrors, &(stats->txerr),
			   &last, delta);
	    fmt_counter_dt(s, apwirelessTable[i]->wirelessNumRXErrors, &(stats->rxerr),
			   &last, delta);

	    g_string_append(s, "\n");
    }
  }
    
  last = now;
  done:
    if (apwirelessTable)
      airport_basestation_3_mib_free_wirelessClientsTable(apwirelessTable);

  return code;
}

static int
show_airport_dhcp(scli_interp_t *interp, int argc, char **argv)
{
  airport_basestation_3_mib_dhcpClient_t **apdhcpTable = NULL;
  int i, code = SCLI_OK, num;

  g_return_val_if_fail(interp, SCLI_ERROR);
  
  if (argc > 1) {
    return SCLI_SYNTAX_NUMARGS;
  }
  
  if (scli_interp_dry(interp)) {
    return SCLI_OK;
  }

  num = get_dhcp(interp);
  
  if (num == -1) {
    return SCLI_ERROR;
  }

  if (num == 0) {
    g_string_sprintfa(interp->result, "No dhcp clients\n");
    return SCLI_OK;
  }
  
  airport_basestation_3_mib_get_dhcpClientsTable(interp->peer, &apdhcpTable, 0);

  if (interp->peer->error_status) {
    code = SCLI_SNMP;
    goto done;
  }
  
  if (apdhcpTable) {
    g_string_sprintfa(interp->header, "%17s %15s LEASE", "MAC", "IP");
    char *m;

    for (i = 0; apdhcpTable[i]; i++) {
     /*
      printf("%.*s", (int)apdhcpTable[i]->_dhcpClientIDLength, apdhcpTable[i]->dhcpClientID);
      */
      
      m = mac_to_str(apdhcpTable[i]->dhcpMacAddress);
      g_string_sprintfa(interp->result, "%17s ", m);
      g_free(m);

      g_string_sprintfa(interp->result, "%15s ", fmt_ipv4_address(
          apdhcpTable[i]->dhcpIpAddress, SCLI_FMT_ADDR));

      g_string_sprintfa(interp->result, "%8d",
                      *apdhcpTable[i]->dhcpLeaseTime);
      g_string_append(interp->result, "\n");
    }
  } else {
    code = SCLI_ERROR;
  }
  
  done:
    if (apdhcpTable)
      airport_basestation_3_mib_free_dhcpClientsTable(apdhcpTable);
    
  return code;
}

static int
airport_info(scli_interp_t *interp, int argc, char **argv)
{
  int dhcp, wireless;
  airport_basestation_3_mib_abs3SysConf_t *sysconf = NULL;

  if (argc > 1) {
    return SCLI_SYNTAX_NUMARGS;
  }
  
  if (scli_interp_dry(interp)) {
    return SCLI_OK;
  }

  dhcp = get_dhcp(interp);
  if (dhcp == -1){
    return SCLI_ERROR;
  }
  wireless = get_wireless(interp);
  if (wireless == -1) {
    return SCLI_ERROR;
  }

  airport_basestation_3_mib_get_abs3SysConf(interp->peer, &sysconf, 0);
  
  g_string_sprintfa(interp->result, "Wireless clients: %d\n", wireless);
  g_string_sprintfa(interp->result, "DHCP clients:     %d\n", dhcp);
  g_string_sprintfa(interp->result, "Firmware version: %.*s\n", 
        (int)sysconf->_sysConfFirmwareVersionLength,
        sysconf->sysConfFirmwareVersion
        );

  if (sysconf)
    airport_basestation_3_mib_free_abs3SysConf(sysconf);
  
  return SCLI_OK;
}

void
scli_init_airport_mode(scli_interp_t *interp)
{
    static scli_cmd_t cmds[] = {
	{ "show airport info", NULL,
	  "The `show airport info' command shows some basic info\n"
	  "about the airport.",
	  SCLI_CMD_FLAG_NEED_PEER | SCLI_CMD_FLAG_DRY,
	  NULL, NULL,
	  airport_info },

	{ "show airport dhcp", NULL,
	  "The `show airport dhcp' command lists the current\n"
	  "dhcp leases the airport is managing.",
	  SCLI_CMD_FLAG_NEED_PEER | SCLI_CMD_FLAG_DRY,
	  NULL, NULL,
	  show_airport_dhcp },

	{ "show airport wireless", NULL,
	  "The `show airport wireless' command lists the current\n"
	  "wireless nodes associated with the airport.",
	  SCLI_CMD_FLAG_NEED_PEER | SCLI_CMD_FLAG_DRY,
	  NULL, NULL,
	  show_airport_wireless },

	{ "monitor airport wireless", NULL,
	  "The `monitor airport wireless' command monitors\n"
	  "the state of the wireless clients",
	  SCLI_CMD_FLAG_NEED_PEER | SCLI_CMD_FLAG_MONITOR | SCLI_CMD_FLAG_DRY,
	  NULL, NULL,
	  monitor_airport_wireless },

	{ "show airport interfaces", NULL,
	  "The `show airport interfaces' command prints\n"
	  "the status of the airports physical interfaces.",
	  SCLI_CMD_FLAG_NEED_PEER | SCLI_CMD_FLAG_DRY,
	  NULL, NULL,
	  show_airport_physical_interfaces_info },

	{ NULL, NULL, NULL, 0, NULL, NULL, NULL }
    };

    static scli_mode_t airport_mode = {
	"airport",
	"The airport mode shows information about the current status of Apple\n"
	"Airport Wifi Access points\n"
	"It is based on the AIRPORT-BASESTATION-3-MIB (200301160001Z)",
	cmds
    };

    scli_register_mode(interp, &airport_mode);
}
